package   rsf.operator;

import    jam.vector.*;
import    jam.operator.*;
import    rsf.vector.*;

/**
 * Projection of two or three dimensional data to a 1-D line. 
 * (I may need at some point a projection of a three dimensional data to 2-D.) 
 *
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @author    Matt Schwab
 */
public class Projection extends sep.operator.SepLinearOperator {
  protected float[] steps; 
  protected Rsf   weights;
  /**
   * @param domain: 2-D space of the projected data.
   * @param range:  1-D space into which the domain is projected.
   * @param steps:  The vector n=(n_x,n_y, ..) in f(nx) 
   *                (points in the gradient direction).
   */
  public Projection(RsfSpace domain, RsfSpace range) {
    super(domain, range); 
    this.steps   = new float[domain.getDimension()]; 
    this.weights = (Rsf) range.newMember();
    this.weights.fill(1.f); // default 
  }
  public Projection(RsfSpace domain) { 
    this(domain, Projection.createRange(domain));
  }
  // copy constructor
  protected Projection(Projection src) {
    super(src); 
    this.steps   = new float[src.steps.length];
    System.arraycopy(src.steps, 0, this.steps, 0, steps.length); 
    this.weights = (Rsf) src.weights.copy();
  }

  /** This defaults the delta and offset of projection space. */ 
  public static RsfSpace createRange(RsfSpace domain) {
    int length = 0; 
    for (int i=0; i < domain.getDimension(); i++) 
      length += 2 * Math.pow(domain.getAxis(i).nsamples,2.); 
    return new RsfSpace((int) Math.sqrt(length)); 
  }

  /** 
   * alpha.length = 1 Projects a plane.  Invokes setNormal(alpha[0]).
   * alpha.length = 2 Projects a volume. Invokes setNormal(alpha[0],alpha[1]).
   */
  public void setNormal(double[] alpha) {
    switch (alpha.length) {
    case 1: 
      setNormal(alpha[0]);
      break;
    case 2:
      setNormal(alpha[0],alpha[1]);
      break;
    default:
      System.err.println("Error: implemented for 2D and 3D domain only"); 
    }   
  }
  /** 
   * This specifies the direction of the projection of planes. 
   * alpha is the angle with the z-Axis (as usual in geophysics). 
   * (I probably should change this: angle with the x-Axis).
   * This parameterisation is standard for polar coordinates.<BR>
   * <B>n</B> = [sin(a), cos(a)]<BR>
   * Alpha is in mulitples of Math.PI. The units of n are pixels. 
   */
  public void setNormal(double alpha) {
    if (this.steps.length != 2) 
      System.err.println("Error: 2-D dimensional mismatch.");
    this.steps[0] = (float) Math.sin(alpha);
    this.steps[1] = (float) Math.cos(alpha);
    //System.out.println("STEEEEEEEEEPs " + steps[0] + " " + steps[1]);
  }
  /** 
   * This specifies the direction of the projection of volumes. 
   * This can be done 
   * several ways. I choose to define the normal vector by its two 
   * spherical angles. The normal vector <B>n</B> relates to the 
   * one-dimensional source function f() by f(<B>n x</B>). 
   * I specify the normal vector <B>n</B> with two angles. 
   * @param gamma (short g) angle of <B>n</B> with the z axis. 
   * @param alpha (short a) angle of <B>n</B>'s projection into 
   *        the xy plane with the x axis. 
   * This parameterisation is standard for spherical coordinates.<BR>
   * <B>n</B> = [sin(g) cos(a), sin(g) sin(a), cos(g)]<BR>
   * The units are pixels. 
   * <P> 
   * Alternatively, I was considering a definition by dip and strike angle:
   * The strike angle s is the angle between the constant plane and the 
   * x-axis in the x-y plane. The dip angle is the angle between the constant
   * plane and the z-axis. In terms of the gamma and alpha of the spherical 
   * coordinates (normal description), the dip and strike vectors are:
   * strike = [sin(a), -cos(a), 0] 
   * dip    = [-cos(g) cos(a), -cos(g) sin(a), sin(g)]
   */ 
  public void setNormal(double alpha, double gamma) {
    // The 2nd and 0th component should switch when I change Rsf indexing. 
    if (this.steps.length != 3) 
      System.err.println("Error: 3-D dimensional mismatch.");
    this.steps[0] = (float) (Math.sin(gamma) * Math.cos(alpha)); 
    this.steps[1] = (float) (Math.sin(gamma) * Math.sin(alpha));
    this.steps[2] = (float)  Math.cos(gamma) ;                   // fastest ax 
  }

  /** 
   * steps lists the components of the normal vector n in f(nx). 
   * @param steps = {n_x, n_y, n_z} (or {n_x, n_y} for 2-D). 
   * I spread f(x) into the volume by looping over location y[ix][iy][iz]
   * and computing f(y * n) where * is the dot product between the location 
   * vector and the here specified normal vector n.
   * The input array steps is normalized before using it. 
   */  
  public void setStepout(float[] steps) {
    if (this.steps.length != steps.length) 
      System.err.println("Error: stepout dimensional mismatch.");
    float norm = 0.f;
    for (int i=0; i<steps.length; i++) 
      norm += steps[i] * steps[i];
    norm = (float) Math.sqrt(norm); 
    if (norm == 0.f) 
      System.err.println("Error: stepout cannot be of zero length.");
    for (int i=0; i<steps.length; i++) 
      steps[i] = steps[i] / norm;
    System.arraycopy(steps, 0, this.steps, 0, steps.length); 
  }

  public void apply(boolean add, Vector vol, Vector lne) {
    if (((Rsf) lne).getDimension() != 1) 
      System.err.println("Error: this projects to a 1D vector only.");
    //    normalizeStepout(); 
    switch (isFwd ? 
      ((RsfSpace) getDomain()).getDimension() : 
      ((RsfSpace) getRange( )).getDimension() ) {
    case 2: 
      apply2d(add, vol, lne);
      break;
    case 3: 
      apply3d(add, vol, lne); 
      break;
    default: 
      System.err.println("Error: implemented for 2D and 3D domain only"); 
    }
  }

  void apply2d(boolean add, Vector vol, Vector lne) {
    Rsf1 l = (Rsf1) lne; float[]   line = l.d;
    Rsf2 v = (Rsf2) vol; float[][] volm = v.d;
    int n0 = v.getAxis(0).nsamples; 
    int n1 = v.getAxis(1).nsamples; 
    float[] wt = ((Rsf1) weights).d;
    if (! add) { if (isFwd) lne.zero(); else vol.zero(); }
    for (int i1=0; i1 < n1; i1++) { float r1 = i1 * steps[0]; // RSF INDEX SHT
    for (int i0=0; i0 < n0; i0++) { float r0 = i0 * steps[1]; // RSF INDEX SHT
      float jx = Math.abs(r1+r0);
      int   j0 = (int) jx;
      float jd = -j0 + jx ;
      if(isFwd) {
	line[j0]     += wt[j0  ] * (1-jd) * volm[i1][i0];
	line[j0+1]   += wt[j0+1] *    jd  * volm[i1][i0];
      }      
      else {
	volm[i1][i0] += wt[j0  ] * (1-jd) * line[j0  ]  ;
	volm[i1][i0] += wt[j0+1] *    jd  * line[j0+1]  ;
      }
    }}
  }
  void apply3d(boolean add, Vector vol, Vector lne) {
    Rsf1 l = (Rsf1) lne; float[]     line = l.d;
    Rsf3 v = (Rsf3) vol; float[][][] volm = v.d;
    int n0 = v.getAxis(0).nsamples; 
    int n1 = v.getAxis(1).nsamples; 
    int n2 = v.getAxis(2).nsamples; 
    float[] wt = ((Rsf1) weights).d;
    if (! add) { if (isFwd) lne.zero(); else vol.zero(); }
    for (int i2=0; i2 < n2; i2++) { float r2 = i2 * steps[0]; // RSF INDEX SHT
    for (int i1=0; i1 < n1; i1++) { float r1 = i1 * steps[1]; // RSF INDEX SHT
    for (int i0=0; i0 < n0; i0++) { float r0 = i0 * steps[2]; // RSF INDEX SHT
    float jx = Math.abs(r2 + r1 + r0);
      int   j0 = (int) jx;
      float jd = -j0 + jx;
      if(isFwd) {
	line[j0]         += wt[j0  ] * (1-jd) * volm[i2][i1][i0];
	line[j0+1]       += wt[j0+1] *    jd  * volm[i2][i1][i0];
      }      
      else {
	volm[i2][i1][i0] += wt[j0  ] * (1-jd) * line[j0  ]  ;
	volm[i2][i1][i0] += wt[j0+1] *    jd  * line[j0+1]  ;
      }
    }}}
  }

  /**
   * Normalizes the forward and adjoint operation so that 
   * adjoint(forward(const)) = const. The scaling is 
   * Sqrt(1 / Number_of_contributions_to_Line_Pixel).<BR>  
   *  
   * HACK!! Reimplements part of the actual operator. 
   * I could compute the analytic solution of the integral
   */
  public void setWeights() {
    switch(((RsfSpace) getDomain()).getDimension()) { 
    case 2: 
      setWeight2d();
      break;
    case 3: 
      setWeight3d(); 
      break;
    default: 
      System.err.println("Error: implemented for 2D and 3D domain only"); 
    }
  }

  void setWeight2d() {
    Rsf1 wt = (Rsf1) weights;                 float[]   line = wt.d;
    Rsf2 vv = (Rsf2) getDomain().newMember(); float[][] volm = vv.d;
    int  n0 = vv.getAxis(0).nsamples; 
    int  n1 = vv.getAxis(1).nsamples; 
    vv.fill(1.f); wt.zero();
    for (int i1=0; i1 < n1; i1++) { float r1 = i1 * steps[0]; // RSF INDEX SHT
    for (int i0=0; i0 < n0; i0++) { float r0 = i0 * steps[1]; // RSF INDEX SHT
      float jx = Math.abs(r1+r0);
      int   j0 = (int) jx;
      float jd = -j0 + jx;
      line[j0]     += (1-jd) * volm[i1][i0];
      line[j0+1]   +=    jd  * volm[i1][i0];
    }}
    ElwSqrt.elwSqrt(weights);
    ElwInv.elwInvDivByZeroIsZero(weights);
  }
  void setWeight3d() {
    Rsf1 wt = (Rsf1) weights;                 float[]     line = wt.d;
    Rsf3 vv = (Rsf3) getDomain().newMember(); float[][][] volm = vv.d;
    int  n0 = vv.getAxis(0).nsamples; 
    int  n1 = vv.getAxis(1).nsamples; 
    int  n2 = vv.getAxis(2).nsamples; 
    vv.fill(1.f); wt.zero();
    for (int i2=0; i2 < n2; i2++) { float r2 = i2 * steps[1]; // RSF INDEX SHT
    for (int i1=0; i1 < n1; i1++) { float r1 = i1 * steps[1]; // RSF INDEX SHT
    for (int i0=0; i0 < n0; i0++) { float r0 = i0 * steps[2]; // RSF INDEX SHT
      float jx = Math.abs(r2+r1+r0);
      int   j0 = (int) jx;
      float jd = -j0 + jx;
      line[j0]     += (1-jd) * volm[i2][i1][i0];
      line[j0+1]   +=    jd  * volm[i2][i1][i0];
    }}}
    ElwSqrt.elwSqrt(weights);
    ElwInv.elwInvDivByZeroIsZero(weights);
  }

  public boolean equals(Object p2) {
    if (this == p2) return true;
    if (! (p2 instanceof Projection)) return false;
    return (super.equals(p2) &&
	    juice.util.JuiceUtil.equalArrays(steps,((Projection) p2).steps) && 
	    weights.equals(((Projection) p2).weights)
	    );
  }

  public jam.JamObject copy() { return new Projection(this); }

  public String toString() {
    return ("Projection (" + (isFwd ? "forward":"adjoint") + "):\n" +
	    "Domain:\n" + getDomain() + "\n" + 
	    "Range: \n" + getRange());
  }
}


