package  rsf.solver;

import    jam.operator.*;
import    jam.vector.*;
import    jam.solver.*;
import  juice.operator.CompoundLinOpAdj;
import    rsf.vector.*;

/**
 * <TT>MaskedEqns</TT> sets up the missing data problem: <BR>
 * A(u + k) = 0                   <BR>
 * That is, given known data, k, we fill in values of u which makes A(u + k)
 * as close to zero as possible.  <BR>
 * So as not to get the trivial solution u = -k, we restrict the search space
 * of u using a masking operator.  By using the default masking operator, the
 * search space is the space in which the non-zero components of k are clamped
 * to zero.
 * <BR><BR>
 * The actual form of the equation is:<BR>
 * AMu = -Ak          <BR>
 * M is the free-data masking operator, k is the known data,
 * and u is the known data for which we solve.
 * <BR><BR>
 * Use getAM() and getNegAk() to retrieve the operator AM and the Vector -Ak.
 * <BR>
 * It is important to make sure that initially u = Mu, so if there is any
 * doubt, use zeroKnownComponents(u).
 *
 * <BR><!--#include file="./source.inc"--><BR>
 *
 */
public class MaskedEqns {
  Mask                             freeMask;
  CompoundLinearOperatorAdjoint    AM;
  Vector                           negAk;

  /**
   * Assumes that the unknown data occurs whereever knownData is zero. <BR>
   * If you need to specify that the data is known to be zero 
   * at certain places, use the constructor which specifies your own mask.
   */
  public MaskedEqns(Operator A, Rsf knownData) {
    this(A, knownData, new SepMask(knownData, true));
  }

  /**
   * This constructor allows known values to be zero by letting you specify the
   * "freeMask", which zeroes all the components of an Rsf which
   * correspond to known data components.
   */
  public MaskedEqns(Operator A, Vector knownData, Mask freeMask) {
    this.freeMask = freeMask;
    negAk = A.image(knownData);
    negAk.neg();
    AM = new CompoundLinOpAdj(freeMask, (hasAdjoint) A);
  }

  /** @return A(M) (M is the operator that zeroes the known data components) */
  public CompoundLinearOperatorAdjoint getAM()    { return AM; }

  /** @return -Ak  (k is the known data) */
  public Vector                        getNegAk() { return negAk; }

  /** sets u = Mu  (M zeroes the known data components) */
  public void                          zeroKnownComponents(Vector u) {
    freeMask.image(u, u);
  }
}
