package rsf.vector;

import   jam.vector.*;
import juice.util.JuiceUtil;
import   jam.JamObject;

/**
 * vector space containing <TT>Rsf</TT>s
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @see       Rsf
 * @author    Joel Schroeder
 */
public class RsfSpace implements SpaceBasis {
  public static final int maxDimension = 6;
  DataAxis dataAxis; 
  Axis[]       axes; 
  
  // constructors
  public RsfSpace() {}

  /** copy constructor */
  public RsfSpace(RsfSpace src) {
    this.dataAxis = (DataAxis) src.dataAxis.copy();
    int n = src.axes.length;
    this.axes = new Axis[n];
    for (int i = 0; i < n; i++)
      this.axes[i] = (Axis) src.axes[i].copy();
  }

  protected RsfSpace(rsfF.vector.RsfSpace src) {
    dataAxis = src.getDataAxis();
    axes     = src.getAxes();
  }
  /** nSamples[0] is fastest axis. It should not. See ../paper/README. */
  public RsfSpace(int [] nSamples ) {
    int n = nSamples.length;
    if (n > maxDimension) {
      System.err.println("Error in RsfSpace constructor: nSamples.length = " +
			 n + " > " + maxDimension);
      return;
    }
    this.dataAxis = new DataAxis();
    this.axes     = new     Axis[n];
    for (int i = 0; i < n; i++ )
      this.axes[i] = new Axis(nSamples[i]);
  }
  public RsfSpace(int    n) {
    this.dataAxis = new DataAxis();
    this.axes     = new     Axis[1];
    this.axes[0]  = new     Axis(n); }
  /** n0 is the fastest index: array[n1][n0] but should not: ../paper/README */
  public RsfSpace(int n0, int n1) {
    this.dataAxis = new DataAxis();
    this.axes    = new Axis[2];
    this.axes[0] = new Axis(n0);
    this.axes[1] = new Axis(n1); }
  /** n0 is the fastest index: array[n2][n1][n0] */
  public RsfSpace(int n0, int n1, int n2) {
    this.dataAxis = new DataAxis();
    this.axes    = new Axis[3];
    this.axes[0] = new Axis(n0);
    this.axes[1] = new Axis(n1);
    this.axes[2] = new Axis(n2); }
  public RsfSpace(Axis[] axes               ) { this(axes, new DataAxis()); }
  public RsfSpace(Axis[] axes, DataAxis axis) {
    this.dataAxis = axis;
    int n = axes.length;
    if (n > maxDimension) {
      System.err.println("Error in RsfSpace constructor: axes.length = " + n +
			 " > " + maxDimension);
      return;
    }
    this.axes = axes;
  }

  public JamObject copy() { return new RsfSpace(this); }
  
  // comparitors
  public boolean equals(Object spc2 ) {
    if (this == spc2)                                            return true;
    if (!(spc2 instanceof RsfSpace ))                            return false; 
    if (!(JuiceUtil.equalArrays(axes, ((RsfSpace) spc2).axes ))) return false;
    if (!(dataAxis.equals(((RsfSpace) spc2).dataAxis)))          return false;
    return true; 
  }
  
  public DataAxis getDataAxis()      { return dataAxis; }
  /** depreciated: use getAxis(int i) since better encapsulation */
  public Axis[]   getAxes    ()      { return axes;     }
  /** getAxis(0) is fastest axis */
  public Axis     getAxis    (int i) { return axes[i];  }
  
  public Vector  newMember()         { return RsfFactory.newRsf(this); }
  public boolean hasMember(Vector v) { return this.equals(v.getSpace()); }
  public OrthonormalBasis getBasis() { return new RsfBasis(this); }
  
  /**
   * Return a new RsfSpace which is the Cartesian product of two RsfSpaces.<BR>
   * This is similar to creating a <tt>ProductSpace</tt>, but the result of
   * <tt>cross</tt> is a single RsfSpace instead of a <tt>ProductSpace</tt>.
   * 
   * @see jam.vector.ProductSpace
   */
  public RsfSpace cross(RsfSpace spc2) {
    if ( ! this.dataAxis.equals(spc2.dataAxis)) {
      System.err.println("Error in RsfSpace cross: this.dataAxis" + 
			 " and spc2.dataAxis are not equal");
    }; 
    int d1 = this.getDimension();
    int d2 = spc2.getDimension();
    Axis[] newAxes = new Axis[ d1 + d2 ];
    for (int i = 0; i < d1; i++ ) { newAxes[i]      = this.getAxes()[i]; }
    for (int i = 0; i < d2; i++ ) { newAxes[i + d1] = spc2.getAxes()[i]; }
    return new RsfSpace(newAxes, this.dataAxis);
  }
  
  // accessors
  public int   getDimension ()      { return axes.length; }
  public int   getNSamples  (int i) {
    return (i < getDimension()) ? (axes[i].nsamples):1;
  }
  public int[] getNSamples() {
    int[] dims = new int[axes.length];
    for (int i = 0; i < dims.length; i++)
      dims[i] = axes[i].nsamples;
    return dims;
  }
  /** @return the number of components in the RsfSpace, i.e. the vector dimension */
  public int size() {
    int size = 1;
    for (int i = 0; i < axes.length; i++ )
      size *= getNSamples(i);
    return size;
  }
  
  // I/O
  public String toString() {
    return (dataAxis + "\n" + JuiceUtil.arrayToString("axis", axes, "\n" ) ); 
  }
}

class RsfBasis implements OrthonormalBasis {
  RsfSpace space;

  public RsfBasis(RsfSpace space) { this.space = space; }

  public SpaceBasis getSpace() { return space; }
  public int        size()     { return space.size(); }
  public Vector     getVector(int i) {
    // We rely on the fact? that arrays of floats are initialized to 0
    Rsf r = RsfFactory.newRsf(space);
    System.err.println("Error in RsfBasis.getVector(): " +
		       "not yet implemented.\n");
    //    r.setDatum(i, 1F);
    return r;
  }

  public JamObject  copy  ()   { return new RsfBasis(space); }

  public boolean    equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof RsfBasis)) return false;
    return (this.space == ((RsfBasis) obj2).space);
  }

  public String toString() {
    StringBuffer sb = new StringBuffer("RsfBasis:\n" +
				       "\tSpace:\n"  + space +
				       "\tBasis:");
    for (int i = 0; i < size(); i++)
      sb.append("\n\t\tVector " + i + ":\n" + getVector(i));
    return sb.toString();
  }
}
